/*
								+----------------------------------+
								|                                  |
								|    ***   File to header   ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2009   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include <string>
#include <algorithm>
#include "../UberLame_src/Integer.h"
#include "../UberLame_src/Tga.h"
#include "../UberLame_src/Jpeg.h"
#include "../UberLame_src/MinMax.h"
#include "../UberLame_src/StlUtils.h"
#include "../UberLame_src/Buffer.h"
#include "../UberLame_src/Compress.h"
#include "../UberLame_src/Base64.h"
#include "../UberLame_src/Base85.h"
#include "../UberLame_src/Crc.h"
#include "../UberLame_src/Dir.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for) && _MSC_VER <= 1200
#define for if(0) {} else for
#endif // _MSC_VER && !__MWERKS__ && !for && _MSC_VER <= 1200
// msvc 'for' scoping hack

#define APP_VERSION "v2.03"

bool ReadFile(TBuffer &r_t_output, const char *p_s_filename)
{
	FILE *p_fr;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fr, p_s_filename, "rb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fr = fopen(p_s_filename, "rb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		return false;
	int n_length;
	if(fseek(p_fr, 0, SEEK_END) ||
	   (n_length = ftell(p_fr)) < 0 ||
	   fseek(p_fr, 0, SEEK_SET)) {
		fclose(p_fr);
		return false;
	}
	if(!r_t_output.Resize(n_length, false)) {
		fclose(p_fr);
		return false;
	}
	if(fread(r_t_output.p_Data(), n_length, sizeof(uint8_t), p_fr) != 1) {
		fclose(p_fr);
		return false;
	}
	fclose(p_fr);
	return true;
}

bool ReadFile(std::string &r_s_output, const char *p_s_filename)
{
	FILE *p_fr;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fr, p_s_filename, "rb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fr = fopen(p_s_filename, "rb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		return false;
	int n_length;
	if(fseek(p_fr, 0, SEEK_END) ||
	   (n_length = ftell(p_fr)) < 0 ||
	   fseek(p_fr, 0, SEEK_SET)) {
		fclose(p_fr);
		return false;
	}
	r_s_output.erase();
	if(!stl_ut::Resize_To_N(r_s_output, n_length)) {
		fclose(p_fr);
		return false;
	}
	if(fread(&r_s_output[0], n_length, sizeof(char), p_fr) != 1) {
		fclose(p_fr);
		return false;
	}
	fclose(p_fr);
	return true;
}

void EscapeString(std::string &r_s_str) // throws
{
	for(size_t i = 0, n = r_s_str.length(); i < n; ++ i) {
		if((uint8_t(r_s_str[i]) < ' ' && r_s_str[i] != '\t') || // keep tabs, we like it
		   r_s_str[i] == '\'' || r_s_str[i] == '\"' || r_s_str[i] == '\\') { // need to handle quotes
			char n_seq = 0;
			switch(r_s_str[i]) {
			case '\n':
				n_seq = 'n';
				break;
			case '\r':
				n_seq = 'r';
				break;
			case '\\':
				n_seq = '\\';
				break;
			case '\'':
				n_seq = '\'';
				break;
			case '\"':
				n_seq = '\"';
				break;
			/*case '\t':
				n_seq = 't';
				break;*/ // keep tabs, we like it
			case '\f':
				n_seq = 'f';
				break;
			case '\a':
				n_seq = 'a';
				break;
			case '\b':
				n_seq = 'b';
				break;
			case '\v':
				n_seq = 'v';
				break;
			};
			// see if we can find escape sequence for this character

			if(n_seq) {
				r_s_str[i] = n_seq;
				r_s_str.insert(i, 1, '\\');
				++ i;
				++ n; // !!
			} else {
				char p_s_xseq[64];
				stl_ut::Format(p_s_xseq, sizeof(p_s_xseq), "x%02x", int(r_s_str[i]));
				r_s_str[i] = '\\';
				r_s_str.insert(i + 1, (const char*)p_s_xseq);
				size_t l = strlen(p_s_xseq);
				i += l;
				n += l; // !!
			}
			// escape low characters
		} else if(uint8_t(r_s_str[i]) >= 128) {
			char p_s_xseq[64];
			stl_ut::Format(p_s_xseq, sizeof(p_s_xseq), "x%02x", int(uint8_t(r_s_str[i]))); // make sure it is not negative
			r_s_str[i] = '\\';
			r_s_str.insert(i + 1, (const char*)p_s_xseq);
			size_t l = strlen(p_s_xseq);
			i += l;
			n += l; // !!
			// escape big characters
		}
	}
	// todo - not sure what happens if "\x12" escape sequence is followed by a [0-9a-f] letter?
	// maybe need to escape even simple letters until the next letter is another one
}

// t_odo - skip mutliline comments, actually not that hard, first erase the single line comment,
// then just make sure that there is no "\"" before the "/*" and then erase everything up to the next "*/"
// not checking for the line comments

// note that decoupling comments from printing would make it much easier

/*void Strip_LineComment(std::string &r_s_comment, std::string &r_s_line,
	size_t n_line, size_t n_line_num, uint64_t &r_n_file_size)
{
	r_s_comment.clear();
	// !!

	size_t n_pos;
	if((n_pos = r_s_line.find("//")) != std::string::npos &&
	   (r_s_line.rfind('\"', n_pos) == std::string::npos ||
	   r_s_line.find('\"', n_pos) == std::string::npos) && // not arround strings, seems dangerous (can have strings before or in, but not on both sides)
	   (r_s_line[r_s_line.length() - 1] == '\n' || (r_s_line[r_s_line.length() - 2] == '\\' &&
	   r_s_line[r_s_line.length() - 1] == 'n') || n_line + 1 == n_line_num)) { // must be the whole line here (so that the comment also ends)
		if(r_s_line[r_s_line.length() - 1] == 'n') {
			size_t e = r_s_line.length() - 1, n_slash_num = 0;
			while(e > 0 && r_s_line[e - 1] == '\\') {
				-- e;
				++ n_slash_num;
			}
			_ASSERTE(n_slash_num > 0);
			if(!(n_slash_num & 1))
				return;
		}
		// this can still turn out bad, the end "\n" might be something like "\\n" and
		// the backslash does not belong to the "n" and hence the line does not in fact end in a newline
		// or it can be "\\\n" and then it is ok "\\a\\n" does not get confused - so only need to count
		// the consecutive backslashes and if even, it is not a newline

		size_t n_end = r_s_line.length();
		if(n_end > 1 && r_s_line[n_end - 2] == '\\' && r_s_line[n_end - 1] == 'n')
			n_end -= 2;
		else if(n_end > 1 && r_s_line[n_end - 1] == '\n')
			-- n_end;
		// skip the newline at the end

		r_s_comment.insert(r_s_comment.end(),
			r_s_line.begin() + n_pos,
			r_s_line.begin() + n_end);
		// extract the comment

		while(n_pos > 0 && isspace(uint8_t(r_s_line[n_pos - 1])))
			-- n_pos;
		// skip whitespace before the comment

		size_t n_comment_escape_num = 0;
		for(size_t i = 0, n = r_s_comment.length(); i < n; ++ i) {
			if(r_s_comment[i] == '\\') {
				if(i + 1 < n && r_s_comment[i + 1] == 'x')
					n_comment_escape_num += 3; // e.g. \x01
				else
					++ n_comment_escape_num;
				++ i; // to avoid counting backslash escape seq. twice
			}
		}
		// counts escape sequences in strings (!!)

		r_n_file_size -= n_end - n_pos; // account for it in the file size
		r_n_file_size += n_comment_escape_num; // these were accounted for twice
		r_s_line.erase(r_s_line.begin() + n_pos,
			r_s_line.begin() + n_end);
		// erase the comment from the original string
	}
}*/

#if 0
void Base64_Evolution()
{
	int p_ic_tab[128] = {0};
	/*for(unsigned int i = 0; i < 128; ++ i) {
		p_ic_tab[i] = (isalpha(i))? (i | 32) - 'a' + ((i < 'a')? 0 : 26) :
			(isdigit(i))? i - '0' + 52 : (i == '+')? 62 : 63;
	}*/
	for(int i = 0; i < 64; ++ i)
		p_ic_tab[(i < 26)? 65 + i : (i < 52)? 71 + i : (i < 62)? i - 4 : "+/"[i - 62]] = i;
		//p_ic_tab[(i < 26)? 65 + i : (i < 52)? 71 + i : (i < 62)? i - 4 : (i - 51) * 4 - 1] = i;
		//p_ic_tab[(i < 26)? 65 + i : (i < 52)? 71 + i : (i < 62)? i - 4 : (i > 62)? '/' : '+'] = i;
		//p_ic_tab[(i < 26)? 'A' + i : (i < 52)? 'a' - 26 + i : (i < 62)? '0' - 52 + i : (i == 62)? '+' : '/'] = i;
		//p_ic_tab["ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"[i]] = i;
}
#endif // 0

void PrintHelp()
{
	fprintf(stderr, "error: missing program arguments.\n"
		"usage:\n"
		"FileToHeader <input-file> <header-file> [options]\n"
		"where input-file is source file, containing data in text or binary format,\n"
		"and header-file is output file to contain this data.\n"
		"\nOptions may be the following:\n"
		"--verbose parameter turns on verbose output (default off)\n"
		"--name-after overrides input file name for naming the class (in case the input file\n"
		"             is an intermediate of a longer processing chain and does not have a name)\n"
		"--class-name overrides the default class name (not checked for syntax errors)\n"
		"--binary forces binary encoding with compression (binary data / obfuscation)\n"
		"--text saves the input file as transparent text (default)\n"
		"--max-cstr-len sets maximum \"C\"-string length (default 32767)\n"
		"--max-line-len sets maximum line length in the generated header\n"
		"               (only applies to the lines with data; default 100)\n"
		"\nThe following options only apply in text mode:\n"
		"--keep-newlines transcripts newlines as they are found in the file\n"
		"--unix-newlines forces single \'\\n\' newlines (default)\n"
		"--strip-comments enables processing of C++ line comments (the coments are present\n"
		"                 in the header, but are not part of the string)\n"
		"--line-numbers enables putting line numbers as comments in the output file\n"
		"                 for example to aid debugging without the original file, etc.\n"
		"--trim-line-whitespace enables throwing away line whitespace\n"
		"--discard-empty-lines enables throwing away lines with nothing but whitespace\n"
		"\nThe following options only apply in binary mode:\n"
		"--shared-decompress enables shared decompressor class (disabled by default)\n"
		"--force-text-ops enables text processing operations (disabled by default)\n"
		"--base-64 forces the use of Base-64 instead of the default Base-85\n");
}

#include "Decompress64.h"
#include "Decompress64Src.h"
#include "Decompress85.h"
#include "Decompress85Src.h"

//#include "Test_TextMode3.h" // simple test

void TrimSpace_but_EndLine(std::string &r_s_line)
{
	size_t b = 0, e = r_s_line.length();

	while(e > b && (r_s_line[e - 1] == '\r' || r_s_line[e - 1] == '\n'))
		-- e;
	// keep newline

	size_t e2 = e;
	while(b < e2 && isspace(uint8_t(r_s_line[b])))
		++ b;
	while(b < e2 && isspace(uint8_t(r_s_line[e2 - 1])))
		-- e2;
	// find whitespace

	r_s_line.erase(e2, e - e2);
	r_s_line.erase(0, b);
}

int main(int n_arg_num, const char **p_arg_list)
{
	if(n_arg_num < 3) {
		PrintHelp();
		return 0;
	}

	/*{
		int m = CMain_cpp::n_Size();
		const char *p_s_text = (const char*)CMain_cpp::p_Data();
		int n = strlen(p_s_text);
		if(m == n)
			printf("recorded and actual length (%d) matches\n", m);
		else
			printf("recorded (%d) and actual (%d) length do not match\n", m, n);
		uint32_t n_crc = CCrc_32::n_Crc(n, p_s_text);
		if(n_crc == CMain_cpp::n_CRC32())
			printf("recorded and actual CRC (0x%08x) matches\n", n_crc);
		else
			printf("recorded (0x%08x) and actual (0x%08x) CRC do not match\n", CMain_cpp::n_CRC32(), n_crc);
		FILE *p_fw = fopen("test_writeback.h", "wb");
		fprintf(p_fw, "%s", p_s_text);
		fclose(p_fw);
		if(CMain_cpp::b_Dynamic())
			free((char*)p_s_text);
	}*/ // simple test

	unsigned int n_max_string_length = 32767; // to avoid c++ compilers complaining
	unsigned int n_max_line_length = 100;
	bool b_verbose = false;
	bool b_binary_mode = false;
	bool b_prefix_line_numbers = false; // in text mode, add /* 123 */ line numbers in front of the strings with the lines
	bool b_overriden_newline_mode = false; // if --keep-newlines or --unix-newlines was specified
	bool b_unix_newlines = true; // in text mode, replaces "\r\n" sequences for only "\n"
	bool b_strip_line_comments = false; // in text mode, get "// line comments" and put them outside of the text, effectively reducing the payload
	bool b_shared_decompress = false; // whether to make the decompression routine shared (potentially leads to link conflicts)
	bool b_treat_binary_as_text = false; // force text processing in binary mode as well
	bool b_trim_line_whitespace = false; // throw away leading and trailing whitespace (except the newline) on each line
	bool b_discard_empty_lines = false; // throw away lines that only contain whitespace (including a newline)
	bool b_use_base85 = true; // use Base-85 instead of Base-64 (default)
	const int n_tab_size = 4;
	// config

	const char *p_s_infile = p_arg_list[1];
	const char *p_s_name_file = p_s_infile;
	const char *p_s_outfile = p_arg_list[2];
	const char *p_s_classname = 0;

	for(int i = 3; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "--verbose") || !strcmp(p_arg_list[i], "-v"))
			b_verbose = true;
		else if(!strcmp(p_arg_list[i], "--keep-newlines") || !strcmp(p_arg_list[i], "-kn")) {
			b_unix_newlines = false;
			b_overriden_newline_mode = true;
		} else if(!strcmp(p_arg_list[i], "-h") || !strcmp(p_arg_list[i], "--help")) {
			PrintHelp();
			return 0;
		} else if(!strcmp(p_arg_list[i], "--unix-newlines") || !strcmp(p_arg_list[i], "-un")) {
			b_unix_newlines = true;
			b_overriden_newline_mode = true;
		} else if(!strcmp(p_arg_list[i], "--shared-decompress") || !strcmp(p_arg_list[i], "-sd"))
			b_shared_decompress = true;
		else if(!strcmp(p_arg_list[i], "--line-numbers") || !strcmp(p_arg_list[i], "-ln"))
			b_prefix_line_numbers = true;
		else if(!strcmp(p_arg_list[i], "--strip-comments") || !strcmp(p_arg_list[i], "-sc"))
			b_strip_line_comments = true;
		else if(!strcmp(p_arg_list[i], "--trim-line-whitespace") || !strcmp(p_arg_list[i], "-tlw"))
			b_trim_line_whitespace = true;
		else if(!strcmp(p_arg_list[i], "--discard-empty-lines") || !strcmp(p_arg_list[i], "-del"))
			b_discard_empty_lines = true;
		else if(!strcmp(p_arg_list[i], "--binary") || !strcmp(p_arg_list[i], "-b"))
			b_binary_mode = true;
		else if(!strcmp(p_arg_list[i], "--text") || !strcmp(p_arg_list[i], "-t"))
			b_binary_mode = false;
		else if(!strcmp(p_arg_list[i], "--force-text-ops") || !strcmp(p_arg_list[i], "-fto"))
			b_treat_binary_as_text = true;
		else if(!strcmp(p_arg_list[i], "--base-64") || !strcmp(p_arg_list[i], "-b64"))
			b_use_base85 = false;
		else if(i + 1 == n_arg_num) {
			fprintf(stderr, "warning: parameter: \'%s\': "
				"unknown parameter or needs a value (skipping)\n", p_arg_list[i]);
		} else if(!strcmp(p_arg_list[i], "--max-line-len") || !strcmp(p_arg_list[i], "-mll"))
			n_max_line_length = atol(p_arg_list[++ i]);
		else if(!strcmp(p_arg_list[i], "--max-cstr-len") || !strcmp(p_arg_list[i], "-mcsl"))
			n_max_string_length = atol(p_arg_list[++ i]);
		else if(!strcmp(p_arg_list[i], "--class-name") || !strcmp(p_arg_list[i], "-cn"))
			p_s_classname = p_arg_list[++ i];
		else if(!strcmp(p_arg_list[i], "--name-after") || !strcmp(p_arg_list[i], "-na"))
			p_s_name_file = p_arg_list[++ i];
		else
			fprintf(stderr, "warning: unknown parameter: \'%s\' (skipping)\n", p_arg_list[i]);
	}
	// parse command-line

	if(b_binary_mode) {
		if(!b_treat_binary_as_text) {
			if(b_prefix_line_numbers || b_overriden_newline_mode || b_strip_line_comments ||
			   b_trim_line_whitespace || b_discard_empty_lines) {
				fprintf(stderr, "warning: --force-text-ops not specified, all"
					" text operations ignored in binary mode\n");
			}
		} else {
			if(b_prefix_line_numbers)
				fprintf(stderr, "warning: --line-numbers ignored in binary mode\n");
		}
	} else {
		if(b_treat_binary_as_text)
			fprintf(stderr, "warning: --force-text-ops ignored in text mode\n");
		if(!b_use_base85)
			fprintf(stderr, "warning: --base-64 ignored in text mode\n");
	}
	// check command-line

	if(b_verbose)
		printf("FileToHeader " APP_VERSION " (built " __DATE__ ")\n\n");
	// print version

	n_max_string_length = max(n_max_string_length, n_max_line_length);
	// must be at least the same

	n_max_line_length -= 2; // the quotes
	n_max_line_length -= 3 * n_tab_size; // the tabs

	if(b_verbose)
		printf("loading input file ...\n");

	uint64_t n_file_size = -1, n_orig_file_size = -1;
	bool b_is_dynamic = false;
	uint32_t n_crc32 = 0, n_orig_crc32 = 0;
	std::string s_data;
	int n_base = (b_use_base85)? 85 : 64; // if compressed
	try {
		std::string s_file;
		std::vector<int> line_no;
		char p_s_line_no_format[64];
		char p_s_hide_line_no[64];
		std::vector<std::string> line_list;
		std::vector<std::vector<std::string> > line_comment_list;
		// text processing data (shared now)

		if(!b_binary_mode || b_treat_binary_as_text) {
			if(!ReadFile(s_file, p_s_infile)) {
				fprintf(stderr, "error: failed to load \'%s\'\n", p_s_infile);
				return -1;
			}
			// read

			if(s_file.find('\0') != std::string::npos) {
				fprintf(stderr, "error: file \'%s\' contains null characters;"
					" must be saved as binary\n", p_s_infile);
				return -1;
			}

			n_orig_crc32 = CCrc_32::n_Crc(s_file.length() * sizeof(char), s_file.data());
			// calculate the original CRC32

			n_file_size = s_file.size();
			n_orig_file_size = n_file_size;
			// take size now

			if(b_unix_newlines) {
				size_t n_pos = 0;
				while((n_pos = s_file.find("\r\n", n_pos)) != std::string::npos) {
					s_file.erase(n_pos, 1);
					++ n_pos;
				}
			}
			// take binary newlines '\r\n' and make them only '\n'

			stl_ut::Split(line_list, s_file, "\n");
			for(size_t i = 0, n = line_list.size(); i + 1 < n; ++ i)
				line_list[i] += '\n'; // keep newlines
			// split to lines

			if(b_trim_line_whitespace) {
				for(size_t i = 0, n = line_list.size(); i < n; ++ i)
					TrimSpace_but_EndLine(line_list[i]); // trim space
			}
			// trim line whitespace

			if(b_verbose && b_binary_mode)
				printf("forced text processing | ");
			if(b_verbose)
				printf("have %d lines\n", line_list.size());
			// verbose

			if(line_list.back().empty())
				line_list.erase(line_list.end() - 1);
			// erase the last empty line (without even newline on it), it looks weird

			if(b_prefix_line_numbers) {
				stl_ut::Format(p_s_line_no_format, sizeof(p_s_line_no_format),
					PRIsize, line_list.size());
				size_t n_line_num_log10 = strlen(p_s_line_no_format);
				stl_ut::Format(p_s_line_no_format, sizeof(p_s_line_no_format),
					"/* %%%dd */ ", n_line_num_log10);
				// make format string for line numbers

				char p_s_hide_line_no_format[64];
				stl_ut::Format(p_s_hide_line_no_format, sizeof(p_s_hide_line_no_format),
					"   %%%ds    ", n_line_num_log10); // just spaces
				stl_ut::Format(p_s_hide_line_no, sizeof(p_s_hide_line_no), p_s_hide_line_no_format, "");
				for(int b = 0, e = strlen(p_s_hide_line_no); b < e; ++ b) {
					if(b + n_tab_size < e) {
						p_s_hide_line_no[b] = '\t';
						e -= n_tab_size - 1;
						p_s_hide_line_no[e] = 0;
					}
				}
				// make string for hidden line numbers (retarded but working)

				n_line_num_log10 += 7; // 7 = strlen("/* ") + strlen(" */ ")
				if(n_max_line_length > n_line_num_log10)
					n_max_line_length -= n_line_num_log10;
				// include the line numbers in line length calculation

				line_no.resize(line_list.size());
				for(size_t i = 0, n = line_list.size(); i < n; ++ i)
					line_no[i] = i + 1;
				_ASSERTE(line_no.size() == line_list.size());
				// make a list of line numbers (a stupid approach, but easy to get right)
			}

			if(b_strip_line_comments) {
				line_comment_list.resize(line_list.size());
				bool b_in_multiline_comment = false;
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					std::string &r_s_line = line_list[i];
					// get a line

					_ASSERTE(!r_s_line.empty() && (i + 1 == n || r_s_line[r_s_line.length() - 1] == '\n'));
					// make sure each line ends in newline at this stage

					size_t b = 0, e = r_s_line.length();
					if(b_in_multiline_comment) {
						if((b = r_s_line.find("*/")) != std::string::npos) {
							++ b;
							_ASSERTE(r_s_line[b - 1] == '*' && r_s_line[b] == '/');
							// ignore everything else, just look for the end of the comment

							size_t b2 = 0;
							while(isspace(uint8_t(r_s_line[b2])))
								++ b2;
							// find the beginning of the comment text

							_ASSERTE(b > 0);
							size_t e4 = b - 1;
							while(e4 > b2 && isspace(uint8_t(r_s_line[e4 - 1])))
								-- e4;
							// find the end of the comment text

							std::string s_comment;
							s_comment.insert(s_comment.begin(), r_s_line.begin() + b2, r_s_line.begin() + e4);
							if(!s_comment.empty())
								line_comment_list[i].push_back(s_comment);
							r_s_line.erase(r_s_line.begin(), r_s_line.begin() + (b + 1));
							// process the comment

							b = 0; // !!
							e = r_s_line.length(); // this changed
							b_in_multiline_comment = false; // !!
						} else { // the comment is this whole line
							b = 0;
							size_t e2 = e;
							while(e2 > b && (r_s_line[e2 - 1] == '\n' || r_s_line[e2 - 1] == '\r'))
								-- e2;
							// find end of line

							size_t b2 = 0;
							while(b2 < e2 && isspace(uint8_t(r_s_line[b2])))
								++ b2;
							// find the beginning of the comment text

							size_t e4 = e2;
							while(e4 > b2 && isspace(uint8_t(r_s_line[e4 - 1])))
								-- e4;
							// find end of the comment text

							std::string s_comment;
							s_comment.insert(s_comment.begin(), r_s_line.begin() + b2, r_s_line.begin() + e4);
							if(!s_comment.empty())
								line_comment_list[i].push_back(s_comment);
							r_s_line.erase(r_s_line.begin(), r_s_line.begin() + e2);
							// process the comment

							e = b;
							// nothing else to do on this line
						}
					}
					// parse continuing comment

					for(; b < e; ++ b) {
						if(r_s_line[b] == '\'') {
							++ b;
							for(; b < e; ++ b) {
								if(r_s_line[b] == '\\') {
									++ b;
									// can be whatever after it (mostly a single character,
									// sometimes hexadecimal number), but will correctly skip "\'"
								} else if(r_s_line[b] == '\'')
									break;
								// found the other (unescaped) apostrophe
							}
							if(b == e || r_s_line[b] == '\'')
								continue;
						}
						// eat "C" character literals

						if(r_s_line[b] == '\"') {
							++ b;
							for(; b < e; ++ b) {
								if(r_s_line[b] == '\\') {
									++ b;
									// can be whatever after it (mostly a single character,
									// sometimes a hexadecimal number), but will correctly skip "\""
								} else if(r_s_line[b] == '\"')
									break;
								// found the other (unescaped) quote
							}
							if(b == e || r_s_line[b] == '\"')
								continue;
						}
						// eat "C" string literals

						if(r_s_line[b] == '/' && b + 1 < e) {
							if(r_s_line[b + 1] == '/') {
								size_t e2 = e;
								while(e2 > b && (r_s_line[e2 - 1] == '\n' || r_s_line[e2 - 1] == '\r'))
									-- e2;
								// find end of line

								size_t b2 = b + 2;
								while(b2 < e2 && isspace(uint8_t(r_s_line[b2])))
									++ b2;
								// find the beginning of the comment text

								while(b > 0 && isspace(uint8_t(r_s_line[b - 1])))
									-- b;
								// will erase the preceding whitespace as well

								std::string s_comment;
								s_comment.insert(s_comment.begin(), r_s_line.begin() + b2, r_s_line.begin() + e2);
								if(!s_comment.empty())
									line_comment_list[i].push_back(s_comment);
								r_s_line.erase(r_s_line.begin() + b, r_s_line.begin() + e2);
								break;
							}
							// found a line comment

							if(r_s_line[b + 1] == '*') {
								size_t e2 = e;
								while(e2 > b && (r_s_line[e2 - 1] == '\n' || r_s_line[e2 - 1] == '\r'))
									-- e2;
								// find end of line

								size_t b2 = b + 2;
								while(b2 < e2 && isspace(uint8_t(r_s_line[b2])))
									++ b2;
								// find the beginning of the comment text

								size_t e3 = b2 + 1;
								while(e3 < e2 && (r_s_line[e3 - 1] != '*' || r_s_line[e3] != '/'))
									++ e3;
								// find the end of the comment text

								while(b > 0 && isspace(uint8_t(r_s_line[b - 1])))
									-- b;
								// will erase the preceding whitespace as well

								std::string s_comment;
								if(e3 < e2 && r_s_line[e3 - 1] == '*' && r_s_line[e3] == '/') {
									// the comment ends on this line

									_ASSERTE(e3 > 0);
									size_t e4 = e3 - 1;
									while(e4 > b2 && isspace(uint8_t(r_s_line[e4 - 1])))
										-- e4;
									// find end of the comment text

									s_comment.insert(s_comment.begin(), r_s_line.begin() + b2, r_s_line.begin() + e4);
									r_s_line.erase(r_s_line.begin() + b, r_s_line.begin() + e3 + 1);
									// get the comment

									e = r_s_line.length(); // this changed
									-- b; // compensate for ++ b in the outer loop
									// (there can be more separate comments on a single line!)
								} else {
									b_in_multiline_comment = true; // the comment carries to the next line

									size_t e4 = e2;
									while(e4 > b2 && isspace(uint8_t(r_s_line[e4 - 1])))
										-- e4;
									// find end of the comment text

									s_comment.insert(s_comment.begin(), r_s_line.begin() + b2, r_s_line.begin() + e4);
									r_s_line.erase(r_s_line.begin() + b, r_s_line.begin() + e2);
									e = b; // we are finished with this line
								}
								// extract the comment

								if(!s_comment.empty())
									line_comment_list[i].push_back(s_comment);
							}
							// found a multiline comment
						}
						// eat comments
					}
					// parse "C" strings, characters and comments / C++ comments
				}
				// for each line

				for(size_t i = 0, n = line_comment_list.size(); i < n; ++ i) {
					std::vector<std::string> &r_comments = line_comment_list[i];
					for(size_t j = 0, m = r_comments.size(); j < m; ++ j) {
						std::string &r_comment = r_comments[j];
						size_t n_pos = 0;
						while((n_pos = r_comment.find("/*")) != std::string::npos)
							r_comment[++ n_pos] = '+';
						n_pos = 0;
						while((n_pos = r_comment.find("*/")) != std::string::npos)
							r_comment[n_pos] = '+';
						++ n_pos;
					}
				}
				// replace comments inside comments to avoid warnings when building the header
			}
			// extract comments from the text

			if(b_discard_empty_lines) {
				_ASSERTE(line_no.empty() || line_no.size() == line_list.size());
				std::string s_line, s_dummy;
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					s_line = line_list[i];
					// get a copy of the line

					stl_ut::TrimSpace(s_line);
					// trim space

					if(s_line.empty()) {
						if(line_comment_list[i].empty()) {
							line_list.erase(line_list.begin() + i);
							if(!line_no.empty())
								line_no.erase(line_no.begin() + i);
							if(!line_comment_list.empty())
								line_comment_list.erase(line_comment_list.begin() + i);
							-- i;
							-- n;
							// no comments, just delete the line
						} else {
							line_list[i].clear();
							// there is a comment, just keep the empty line as a placeholder
						}
					}
					// discard empty ones
				}
				_ASSERTE(line_no.empty() || line_no.size() == line_list.size());
			}
			// discard empty lines

			{
				n_crc32 = CCrc_32::n_Start();
				n_file_size = 0;
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					n_crc32 = CCrc_32::n_Crc(line_list[i].length() * sizeof(char), line_list[i].data(), n_crc32);
					n_file_size += line_list[i].length();
				}
				n_crc32 = CCrc_32::n_Finalize(n_crc32);
				// calculate CRC
			}
			// count file size after all the changes

			if(b_verbose && b_binary_mode) {
				printf("saved " PRIsizeB "B in text processing\n",
					PRIsizeBparams(n_orig_file_size - n_file_size));
			}
		}
		if(!b_binary_mode) {
			std::for_each(line_list.begin(), line_list.end(), EscapeString);
			// escape strange characters

			uint64_t n_print_size = 0;
			for(size_t i = 0, n = line_list.size(); i < n; ++ i)
				n_print_size += line_list[i].length();
			// count length

			if(n_max_line_length > 0) {
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					size_t n_tab_num = std::count(line_list[i].begin(), line_list[i].end(), '\t'); // may not compile in msvc60
					if(line_list[i].length() + n_tab_num * (n_tab_size - 1) > n_max_line_length) { // * (n_tab_size - 1) because one character is counted already
						std::string &r_s_line_i = line_list[i];
						int n_cut;
						unsigned int l = r_s_line_i.length(), chars = 0;;
						for(n_cut = 0; unsigned(n_cut) < l; ++ n_cut) {
							unsigned int chars2 = chars + ((r_s_line_i[n_cut] == '\t')? n_tab_size : 1);
							if(chars2 > n_max_line_length && (chars > n_max_line_length ||
							   (n_max_line_length - chars) < (chars2 - n_max_line_length + 1)))
								break; // find the precise position of the cut, minimizing the distance from the cut line
							chars = chars2;
						}
						/*int n_cut = n_max_line_length - n_tab_num * (n_tab_size - 1); // try to fixup (not precise)*/
						if(n_cut <= 0)
							n_cut = n_max_line_length; // ...
						if(n_cut >= r_s_line_i.length())
							n_cut = r_s_line_i.length(); // 2016-06-16 - fixed a bug with out of bounds access triggered by lines with many tabs on them where the imprecise estimate overshoots past the length of the line
						_ASSERTE(n_cut > 0);
						// decide where to cut to have no more than n_max_line_length characters in the header

						while((n_cut > 0 && r_s_line_i[n_cut - 1] == '\\') ||
						   (n_cut > 1 && r_s_line_i[n_cut - 2] == '\\' && r_s_line_i[n_cut - 1] == 'x') ||
						   (n_cut > 2 && r_s_line_i[n_cut - 3] == '\\' && r_s_line_i[n_cut - 2] == 'x'))
						   n_cut -= (r_s_line_i[n_cut - 1] == '\\')? 1 : (r_s_line_i[n_cut - 1] == 'x')? 2 : 3;
						// can't leave '\\' at the end of the line (shouldn't cut escape sequences in two parts)
						// note that '\\x12' will still be cut by this (the compilers do not like that) // t_odo - test it

						if(n_cut <= 0 || n_cut == r_s_line_i.length())
							continue;
						// leave long line, couldn't cut because of abuse of escape sequences

						std::string s_second_line;
						s_second_line.insert(s_second_line.end(),
							line_list[i].begin() + n_cut, line_list[i].end());
						line_list[i].erase(line_list[i].begin() + n_cut, line_list[i].end());
						line_list.insert(line_list.begin() + (i + 1), s_second_line);
						++ n; // !!

						if(!line_no.empty())
							line_no.insert(line_no.begin() + i + 1, -1); // insert -1 *after* so the lines do not repeat
						if(!line_comment_list.empty())
							line_comment_list.insert(line_comment_list.begin() + i, std::vector<std::string>()); // insert empty comment *before* as there will likely be more space on the subsequent lines
						// duplicate line numbers and comments as well, if we need them
					}
				}
				_ASSERTE(line_no.empty() || line_no.size() == line_list.size());
				_ASSERTE(line_comment_list.empty() || line_comment_list.size() == line_comment_list.size());
			}
			// break long lines to multiple ones (but without introducing new line breaks)

			if(n_print_size > n_max_string_length) { // note that n_print_size could be n_file_size as well, but some compilers might actually limit that instead
				s_data = "\t\tconst char *p_text_array[] = {\n";
				size_t n_cur_size = 0;
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					size_t n_last_line = line_list[i].length();
					if(n_cur_size + n_last_line <= n_max_string_length) {
						if(i)
							s_data += "\n";
						n_cur_size += n_last_line;
					} else {
						if(i)
							s_data += ",\n";
						n_cur_size = n_last_line;
					}
					s_data += "\t\t\t";
					if(b_prefix_line_numbers) {
						char p_s_line_no[64];
						if(line_no[i] > 0) {
							stl_ut::Format(p_s_line_no, sizeof(p_s_line_no),
								p_s_line_no_format, line_no[i]);
							s_data += p_s_line_no;
						} else
							s_data += p_s_hide_line_no;
					}
					s_data += "\"";
					s_data += line_list[i];
					s_data += "\"";
					if(!line_comment_list.empty() && !line_comment_list[i].empty()) {
						for(size_t j = 0, m = line_comment_list[i].size(); j < m; ++ j) {
							s_data += " /* ";
							s_data += line_comment_list[i][j];
							s_data += " */";
						}
						// insert comments
					}
				}
				s_data += "\n\t\t};\n";
				// need to split it

				std::string s_size;
				if(!stl_ut::Format(s_size, "\t\tchar *p_s_text;\n\t\tif(!(p_s_text = new(std::nothrow) char["
				   PRIu64 "]))\n\t\t\treturn 0;\n", n_file_size + 1))
					throw std::bad_alloc(); // retrhow
				s_data += s_size;

				/*s_data += "\t\t*p_s_text = 0;\n"
						  "\t\tfor(size_t i = 0, n = sizeof(p_text_array) / sizeof(p_text_array[0]); i < n; ++ i)\n"
						  "\t\t\tstrcat(p_s_text, p_text_array[i]);\n"
						  "\t\treturn p_s_text;\n";*/
				// uses strcat; ugly, inefficient

				/*std::string s_concat;
				if(!stl_ut::Format(s_concat, "\t\tchar *p_s_dest = p_s_text;\n"
				   "\t\tfor(size_t i = 0, n = sizeof(p_text_array) / sizeof(p_text_array[0]); i < n; ++ i) {\n"
				   "\t\t\tsize_t n_chunk = (i + 1 < n)? " PRIsize " : " PRIsize ";\n"
				   "\t\t\tmemcpy(p_s_dest, p_text_array[i], n_chunk * sizeof(char));\n" "\t\t\tp_s_dest += n_chunk;\n"
				   "\t\t}\n" "\t\t*p_s_dest = 0;\n" "\t\treturn p_s_text;\n", n_max_string_length,
				   size_t(n_file_size % n_max_string_length)))
					throw std::bad_alloc(); // retrhow
				s_data += s_concat;*/ // can't use as we are deciding on n_print_size instead of n_file_size, it is hard to calculate what n_chunk should be
				// better concat code, using memcpy()

				std::string s_concat;
				if(!stl_ut::Format(s_concat, "\t\tchar *p_s_dest = p_s_text;\n"
				   "\t\tfor(size_t i = 0, n = sizeof(p_text_array) / sizeof(p_text_array[0]); i < n; ++ i) {\n"
				   "\t\t\tsize_t n_chunk = strlen(p_text_array[i]);\n"
				   "\t\t\tmemcpy(p_s_dest, p_text_array[i], n_chunk * sizeof(char));\n" "\t\t\tp_s_dest += n_chunk;\n"
				   "\t\t}\n" "\t\t*p_s_dest = 0;\n" "\t\treturn p_s_text;\n", n_max_string_length,
				   size_t(n_file_size % n_max_string_length)))
					throw std::bad_alloc(); // retrhow
				s_data += s_concat;
				// better concat code, using memcpy() and strlen()

				// need to provide code to concat it back

				b_is_dynamic = true;
				// the caller needs to dispose of the text
			} else {
				s_data = "\t\tconst char *p_s_text =\n";
				for(size_t i = 0, n = line_list.size(); i < n; ++ i) {
					s_data += "\t\t\t";
					if(b_prefix_line_numbers) {
						char p_s_line_no[64];
						if(line_no[i] > 0) {
							stl_ut::Format(p_s_line_no, sizeof(p_s_line_no),
								p_s_line_no_format, line_no[i]);
							s_data += p_s_line_no;
						} else
							s_data += p_s_hide_line_no;
					}
					s_data += "\"";
					s_data += line_list[i];
					if(!line_comment_list.empty() && !line_comment_list[i].empty()) {
						s_data += (i + 1 == n)? "\";" : "\"";

						for(size_t j = 0, m = line_comment_list[i].size(); j < m; ++ j) {
							s_data += " /* ";
							s_data += line_comment_list[i][j];
							s_data += " */";
						}
						// insert comments

						s_data += "\n";
					} else
						s_data += (i + 1 == n)? "\";\n\n" : "\"\n";
				}
				// just print them all

				s_data += "\t\treturn p_s_text;\n";

				b_is_dynamic = false;
				// the caller does not need to dispose of the text
			}
		} else {
			TBuffer t_file;
			if(!b_treat_binary_as_text) {
				if(!ReadFile(t_file, p_s_infile)) {
					fprintf(stderr, "error: failed to load \'%s\'\n", p_s_infile);
					return -1;
				}
				n_file_size = t_file.n_Size();
				n_orig_file_size = n_file_size;
				n_crc32 = CCrc_32::n_Crc(t_file.n_Size(), t_file.p_Data());
				n_orig_crc32 = n_crc32;
				// calculate CRC
			} else {
				s_file.erase();
				for(size_t i = 0, n = line_list.size(); i < n; ++ i)
					s_file += line_list[i];
				_ASSERTE(n_file_size == s_file.length() * sizeof(char));
				n_file_size = s_file.length() * sizeof(char);
				if(!t_file.Resize(s_file.length() * sizeof(char)))
					throw std::bad_alloc(); // rehtrow
				memcpy(t_file.p_Data(), s_file.data(), s_file.length() * sizeof(char));
				// just concat; already have a CRC
			}
			// input data

			if(b_verbose)
				printf("compressing ... ");

			TBuffer t_bwt_data;
			uint32_t n_bwt_primary_index;
			{
				if(!CBurrowsWheelerTransform::Encode(t_file, t_bwt_data))
					throw std::bad_alloc(); // rethrow
				_ASSERTE(t_bwt_data.n_Size() == t_file.n_Size() + sizeof(uint32_t));
				n_bwt_primary_index = *(uint32_t*)t_bwt_data.p_Data();
				TBuffer t_tmp;
				if(!t_tmp.Resize(t_file.n_Size()))
					throw std::bad_alloc(); // rethrow
				memcpy(t_tmp.p_Data(), t_bwt_data.p_Data() + sizeof(uint32_t), t_file.n_Size());
				t_tmp.Swap(t_bwt_data);
			}
			// compress the whole file with BWT (do not expect to have very big files in here)

			CMoveToFrontTransform::Encode(t_bwt_data, CMoveToFrontTransform::algo_MTF); // todo - enable this
			// also MTF it, produces shorter files and is easy to decompress

			TBuffer t_packed_data;
			if(!CRunLengthCodec::Encode(t_bwt_data, t_packed_data))
				throw std::bad_alloc(); // rethrow
			// pack using simple RLE (easy to unpack)

			size_t n_afterhuff = t_packed_data.n_Size();

			{
				TBuffer t_tmp;
				if(!CHuffmanCodec::Encode(t_packed_data, t_tmp))
					throw std::bad_alloc(); // rethrow
				t_packed_data.Swap(t_tmp);
			}
			// huffman step

			const char *p_s_base85_alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
				"abcdefghijklmnopqrstuvwxyz!#$%&()*+-;<=>.@^_,{|}~"; // as RFC 1924, but '?' is '.' and '`' is ','
			// modified alphabet for Base-85 that will not cause trigraphs

			std::string s_base64;
			if(b_use_base85) {
				if(!CModifiedBase85::Encode(s_base64, t_packed_data.n_Size() - sizeof(uint32_t),
				   t_packed_data.p_Data() + sizeof(uint32_t), true, p_s_base85_alphabet))
					throw std::bad_alloc(); // rethrow
				// encode as Base-85 with explicit null coding, skip huffman unpack length, we save it otherwise
			} else {
				if(!CModifiedBase64::Encode(s_base64, t_packed_data.n_Size() - sizeof(uint32_t),
				   t_packed_data.p_Data() + sizeof(uint32_t)))
					throw std::bad_alloc(); // rethrow
				// encode as Base-64, skip huffman unpack length, we save it otherwise
			}

			{
				if(b_verbose) {
					printf("packed " PRIu64 " B to " PRIsize " B of Base-%d text\n",
						n_orig_file_size, s_base64.length(), (b_use_base85)? 85 : 64);
					printf("testing decompress ... ");
				}

				uint8_t *p_data;
				if(b_use_base85) {
					p_data = p_Decode_BWT_MTF_RLE_Huff_Base85(s_base64.c_str(),
						size_t(n_file_size), n_afterhuff, n_bwt_primary_index);
				} else {
					p_data = p_Decode_BWT_MTF_RLE_Huff_Base64(s_base64.c_str(),
						size_t(n_file_size), n_afterhuff, n_bwt_primary_index);
				}
				if(!p_data)
					fprintf(stderr, "error: decompress failed\n");
				else if(memcmp(p_data, t_file.p_Data(), size_t(n_file_size)))
					fprintf(stderr, "error: decompress produces different results\n");
				else if(b_verbose)
					printf("decompress works\n");
				if(p_data)
					delete[] p_data;
			}
			// test decompress

			if(b_use_base85) {
				/*size_t n_pos = 0;
				while((n_pos = s_base64.find("??", n_pos)) != std::string::npos) {
					const char *p_s_trigraph_list = "=(</)>!-";
					if(n_pos + 2 < s_base64.length() && strchr(p_s_trigraph_list, s_base64[n_pos + 2]))
						s_base64.insert(n_pos + 1, 1, '\\');
					else
						++ n_pos; // false alarm, not a trigraph
				}*/
				_ASSERTE(s_base64.find("??") == std::string::npos);
				// using a modified alphabet that does not contain question marks
			} else {
				_ASSERTE(s_base64.find("??") == std::string::npos);
				// would not expect trigraphs here, except if using a modified alphabet
			}
			// Base-85 may contain trigraphs, need to escape them properly

			bool b_have_to_free = false;

			size_t n_print_size = s_base64.length();
			if(n_print_size > n_max_string_length) { // note that n_print_size could be n_file_size as well, but some compilers might actually limit that instead
				s_data = "\t\tconst char *p_text_array[] = {\n";
				size_t n_cur_size = 0;
				for(size_t i = 0; i < n_print_size;) {
					size_t n_space = min(size_t(n_max_line_length),
						min(size_t(n_max_string_length) - n_cur_size, n_print_size - i));

					if(n_space) {
						if(i)
							s_data += "\n";
						n_cur_size += n_space;
					} else {
						if(i)
							s_data += ",\n";
						n_cur_size = n_space = min(size_t(n_max_line_length), n_print_size - i);
					}
					s_data += "\t\t\t\"";
					s_data.insert(s_data.end(), s_base64.begin() + i,
						s_base64.begin() + (i + n_space));
					i += n_space;
					s_data += "\"";
				}
				s_data += "\n\t\t};\n";
				// need to split it

				std::string s_size;
				if(!stl_ut::Format(s_size, "\t\tchar *p_s_base%d;\n\t\tif(!(p_s_base%d = new(std::nothrow) char["
				   PRIsize "]))\n\t\t\treturn 0;\n", n_base, n_base, n_print_size + 1))
					throw std::bad_alloc(); // retrhow
				s_data += s_size;;
				/*s_data += "\t\t*p_s_base64 = 0;\n"
						  "\t\tfor(size_t i = 0, n = sizeof(p_text_array) / sizeof(p_text_array[0]); i < n; ++ i)\n"
						  "\t\t\tstrcat(p_s_base64, p_text_array[i]);\n";*/
				// uses strcat; ugly, inefficient

				std::string s_concat;
				if(!stl_ut::Format(s_concat, "\t\tchar *p_s_dest = p_s_base%d;\n"
				   "\t\tfor(size_t i = 0, n = sizeof(p_text_array) / sizeof(p_text_array[0]); i < n; ++ i) {\n"
				   "\t\t\tsize_t n_chunk = (i + 1 < n)? " PRIsize " : " PRIsize ";\n"
				   "\t\t\tmemcpy(p_s_dest, p_text_array[i], n_chunk * sizeof(char));\n" "\t\t\tp_s_dest += n_chunk;\n"
				   "\t\t}\n" "\t\t*p_s_dest = 0;\n", n_base, n_max_string_length,
				   size_t(n_print_size % n_max_string_length)))
					throw std::bad_alloc(); // retrhow
				s_data += s_concat;
				// better concat code, using memcpy
				// need to provide code to concat it back

				b_have_to_free = true;
				// the caller needs to dispose of the text
			} else {
				s_data = (b_use_base85)? "\t\tconst char *p_s_base85 = " : "\t\tconst char *p_s_base64 = ";
				for(size_t i = 0; i < n_print_size;) {
					size_t n_line_length = (i)? n_max_line_length : n_max_line_length -
						strlen("const char *p_s_baseXX = ") + n_tab_size; // n_max_line_length expects three tabs but there are only two
					size_t n_space = min(n_line_length, n_print_size - i);
					s_data += (i)? "\t\t\t\"" : "\"";
					s_data.insert(s_data.end(), s_base64.begin() + i,
						s_base64.begin() + (i + n_space));
					i += n_space;
					s_data += (i == n_print_size)? "\";\n" : "\"\n";
				}
				// just print it all

				b_have_to_free = false;
				// the caller does not need to dispose of the text
			}

			std::string s_decompress;
			if(b_have_to_free) {
				s_data += "\t\tchar *p_data = ";
			} else
				s_data += "\t\treturn ";
			if(b_shared_decompress) {
				if(!stl_ut::Format(s_decompress, "%s(p_s_base%d, " PRIu64 ", " PRIsize ", "
				   PRIsize ");\n", "p_Call_Decompress", n_base, n_file_size, n_afterhuff,
				   size_t(n_bwt_primary_index)))
					throw std::bad_alloc(); // rethrow
			} else {
				if(!stl_ut::Format(s_decompress, "%s%d(p_s_base%d, " PRIu64 ", " PRIsize ", "
				   PRIsize ");\n", "(char*)p_Decode_BWT_MTF_RLE_Huff_Base", n_base, n_base,
				   n_file_size, n_afterhuff, size_t(n_bwt_primary_index)))
					throw std::bad_alloc(); // rethrow
			}
			s_data += s_decompress;
			// prepare the call to the decompress routine

			if(b_have_to_free) {
				if(b_use_base85)
					s_data += "\t\tdelete[] p_s_base85;\n";
				else
					s_data += "\t\tdelete[] p_s_base64;\n";
				s_data += "\t\treturn p_data;\n";
			}
			// return the data

			b_is_dynamic = true;
			// the caller needs to dispose of the text
		}
	} catch(std::bad_alloc&) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// throws

	{
		std::string s_ident, s_header;
		if(!stl_ut::AssignCStr(s_ident, p_s_name_file) ||
		   !stl_ut::AssignCStr(s_header, p_s_name_file)) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		for(size_t i = 0, n = s_ident.length(); i < n; ++ i) {
			if(isalpha(uint8_t(s_ident[i])))
				s_ident[i] = toupper(uint8_t(s_ident[i]));
			else if(!isdigit(uint8_t(s_ident[i]))) {
				s_ident[i] = '_';
				s_header[i] = '_';
			}
		}
		// create class name

		FILE *p_fw;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
		if(fopen_s(&p_fw, p_s_outfile, "w")) {
#else // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		if(!(p_fw = fopen(p_s_outfile, "w"))) {
#endif // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
			fprintf(stderr, "error: failed to open \'%s\' for writing\n", p_s_outfile);
			return -1;
		}

		fprintf(p_fw, "/**\n");
		fprintf(p_fw, " *\t@file %s\n", p_s_outfile);
		fprintf(p_fw, " *\t@brief file in header, generated from \'%s\'\n", p_s_name_file);
		if(b_binary_mode)
			fprintf(p_fw, " *\t\tusing BWT / RLE / Huff / Base%d pipe\n", n_base);
		else
			fprintf(p_fw, " *\t\tstored as plain text\n");
		TFileInfo::TTime t_time = TFileInfo(p_s_name_file).p_time[TFileInfo::time_LastWrite];
		fprintf(p_fw, " *\t@date %4d-%02d-%02d %02d:%02d:%02d\n", t_time.n_year,
			t_time.n_month, t_time.n_day, t_time.n_hour, t_time.n_minute, t_time.n_second);
		fprintf(p_fw, " *\t@author FileToHeader " APP_VERSION " (built " __DATE__ ")\n");
		fprintf(p_fw, " */\n\n");
		fprintf(p_fw, "#pragma once\n");
		fprintf(p_fw, "#ifndef __%s_FILE_IN_HEADER_INCLUDED\n", s_ident.c_str());
		fprintf(p_fw, "#define __%s_FILE_IN_HEADER_INCLUDED\n\n", s_ident.c_str());
		// generate nice header // t_odo - do the same in ImageToHeader

		fprintf(p_fw, "/**\n");
		fprintf(p_fw, " *\t@brief file in header, generated from \'%s\'\n", p_s_name_file);
		fprintf(p_fw, " */\n");
		if(!p_s_classname)
			fprintf(p_fw, "class C%s {\n", s_header.c_str());
		else
			fprintf(p_fw, "class %s {\n", p_s_classname); // class name override
		fprintf(p_fw, "public:\n");
		fprintf(p_fw, "\t/**\n"
					  "\t *\t@brief gets size of the data\n"
					  "\t *\t@return Returns size of the data, in bytes, not including the terminating zero.\n"
					  "\t */\n");
		if(n_orig_file_size == n_file_size) {
			fprintf(p_fw, "\tstatic inline size_t n_Size()\n\t{\n\t\treturn " PRIu64
				"; // matches original file\n\t}\n\n", n_file_size);
		} else {
			fprintf(p_fw, "\tstatic inline size_t n_Size()\n\t{\n\t\treturn " PRIu64
				"; // original file " PRIu64 " B\n\t}\n\n", n_file_size, n_orig_file_size);
		}
		fprintf(p_fw, "\t/**\n"
					  "\t *\t@brief determines whether the stored string is dynamically allocated\n"
					  "\t *\t@return Returns true if the caller of p_Data() needs to free the memory.\n"
					  "\t */\n");
		fprintf(p_fw, "\tstatic inline bool b_Dynamic()\n\t{\n\t\treturn %s;\n\t}\n\n",
			(b_is_dynamic)? "true" : "false");
		fprintf(p_fw, "\t/**\n"
					  "\t *\t@brief gets CRC\n"
					  "\t *\t@return Returns CRC32 of the stored data.\n"
					  "\t *\t@note The stored CRC may differ from the original file if data processing was enabled.\n"
					  "\t */\n");
		if(n_orig_crc32 == n_crc32) {
			fprintf(p_fw, "\tstatic inline uint32_t n_CRC32()\n\t{\n\t\treturn 0x%08x;"
				" // matches original file\n\t}\n\n", n_crc32);
		} else {
			fprintf(p_fw, "\tstatic inline uint32_t n_CRC32()\n\t{\n\t\treturn 0x%08x;"
				" // original file 0x%08x\n\t}\n\n", n_crc32, n_orig_crc32);
		}
		fprintf(p_fw, "\t/**\n"
					  "\t *\t@brief gets data\n"
					  "\t *\n"
					  "\t *\t@return Returns data of the file on success, 0 on failure (not enough memory).\n"
					  "\t *\n"
					  "\t *\t@note The data (despite it may be binary), is additionally null terminated to make\n"
					  "\t *\t\tit easier to use (the terminating null is not counted in n_Size()).\n"
					  "\t *\t@note See also b_Dynamic().\n"
					  "\t */\n");
		if(b_is_dynamic)
			fprintf(p_fw, "\tstatic inline char *p_Data()\n\t{\n");
		else
			fprintf(p_fw, "\tstatic inline const char *p_Data()\n\t{\n");
		fprintf(p_fw, "%s", s_data.c_str());
		fprintf(p_fw, "\t}\n");
		
		uint32_t n_decompressor_crc = 0;
		std::string s_decompress_src;
		if(b_binary_mode) {
			if(b_binary_mode) {
#if 1
				char *p_s_deco;
				if(b_use_base85)
					p_s_deco = CDecompress85_h::p_Data(); // use itself to generate compressed decompressor
				else
					p_s_deco = CDecompress64_h::p_Data(); // use itself to generate compressed decompressor
				if(!p_s_deco || !stl_ut::AssignCStr(s_decompress_src, p_s_deco)) {
					if(p_s_deco)
						delete[] p_s_deco;
					fprintf(stderr, "error: failed to decompress the decompress routines\n");
#else
				if(!ReadFile(s_decompress_src, (b_use_base85)? "Decompress64.h" : "Decompress85.h")) { // uncool, hard to use from commandline
					fprintf(stderr, "error: failed to read \'Decompress.h\',"
						" unable to include decompress routines\n");
#endif
					fclose(p_fw);
					return -1;
				} else {
					size_t n_pos = 0;
					while((n_pos = s_decompress_src.find("\r\n", n_pos)) != std::string::npos)
						s_decompress_src.erase(n_pos, 1);
					// condense newlines

					n_decompressor_crc = CCrc_32::n_Crc(s_decompress_src.length() *
						sizeof(char), s_decompress_src.data());
					// calculate CRC of it
				}
			}
		}
		// read the decompressor

		if(b_binary_mode) {
			if(b_shared_decompress) {
				fprintf(p_fw, "\nprotected:\n"
							  "\t/**\n"
							  "\t *\t@copydoc CFTHDecompressCode_%08x::p_Decode_BWT_MTF_RLE_Huff_Base%d\n"
							  "\t */\n"
							  "\tstatic inline char *p_Call_Decompress(const char *p_s_base%d,\n"
							  "\t\tint n_data_size, int n_huff_stage_size, uint32_t n_bwt_primary_index);\n",
							  n_decompressor_crc, n_base, n_base);
				// specify a small routine that will call decompressor once it's been declared

				fprintf(p_fw, "};\n");
				// end the previos class

				fprintf(p_fw, "\n#ifndef __FILE_TO_HEADER_DECOMPRESS_%08X_INCLUDED\n", n_decompressor_crc);
				fprintf(p_fw, "#define __FILE_TO_HEADER_DECOMPRESS_%08X_INCLUDED\n\n", n_decompressor_crc);

				fprintf(p_fw, "/**\n"
							  " *\t@brief decompression routine implementation\n"
				//			  " *\t@note While this allows all the compressed files to share decompressor implementation,\n"
				//			  " *\t\tit restricts all the files to be included in a single .cpp file, otherwise linking\n"
				//			  " *\t\tfails because of duplicate symbols (remove body of decompress function in such case).\n"
							  " */\n"
							  "class CFTHDecompressCode_%08x {\npublic:\n", n_decompressor_crc);
				fprintf(p_fw, "%s", s_decompress_src.c_str());
				// include decompress routines

				fprintf(p_fw, "};\n\n#endif // __FILE_TO_HEADER_DECOMPRESS_%08X_INCLUDED\n\n", n_decompressor_crc);

				fprintf(p_fw, "inline char *%s%s::p_Call_Decompress(const char *p_s_base%d,\n"
							  "\tint n_data_size, int n_huff_stage_size, uint32_t n_bwt_primary_index)\n"
							  "{\n"
							  "\treturn (char*)CFTHDecompressCode_%08x::p_Decode_BWT_MTF_RLE_Huff_Base%d(p_s_base%d,\n"
							  "\t\tn_data_size, n_huff_stage_size, n_bwt_primary_index);\n"
							  "}\n"
							  "// a dummy function that calls the decompressor, which is now declared\n"
							  "// (allows us to keep it at the end of the file)\n",
							  (p_s_classname)? "" : "C", (p_s_classname)? p_s_classname : s_header.c_str(), // fix: use p_s_classname here as well
							  n_base, n_decompressor_crc, n_base, n_base);
				// implement the body of the dummy routine that calls the decompress after we declared it
			} else {
				fprintf(p_fw, "\nprotected:\n%s};\n", s_decompress_src.c_str());
				// otherwise make the decompressor a part of the file class
			}
		} else
			fprintf(p_fw, "};\n");
		// include decompressor, make it so that the code can be shared if more than a single file is included

		fprintf(p_fw, "\n#endif // !__%s_FILE_IN_HEADER_INCLUDED\n", s_ident.c_str());

		if(ferror(p_fw)) {
			fprintf(stderr, "error: failed to write code to \'%s\'\n", p_s_outfile);
			return -1;
		}

		fclose(p_fw);
	}
	// write C++ header file

	if(b_verbose)
		printf("\ndone\n");

	return 0;
}
